Unit PMan_LFU;
{$R-,S-,O+}

{Project Completion Date : April 7, 1990}

{No modifications to date}

{Uses the Least Frequently Used page replacement algorithm}
{Utilizes an array[NumPages] of Reals, thus requiring a bit}
{more Heap space than PMAN_LRU. Functionally IDENTICAL to}
{PMAN_LRU -- for more detailed comments, see PMAN_LRU}

{Least Frequently Used Page (currently loaded) will be swapped out}
{upon the occurrence of a Page-Fault}

INTERFACE

Uses Pages;

Const
  AbsMaxPageSize = MaxPageSize + SizeOf(Page);
  MaximumSize    = 33554432; {32 MegaBytes}
  Delta          = 0.0000000001;
  MaxPageNum     = MaximumSize Div AbsMaxPageSize;
  MaxFrame       = 426; {MaxMem Div (AbsMaxPageSize + SizeOf(Frame))}

Type
  PathStr    = String[79];
  Frame      = Record
                 PageNum  : LongInt;
                 Current  : Page;
               End;

  FrameArray = Array[0..0] of Frame; {REMEMBER: Frames require 25 Bytes!}
                                     {FrameArray can hold up to 2620 frames}
                                     {but Actual NumFrames will be <= MaxFrame}
  Priorities = Array[0..0] of Real;

  PageManager = Object

                  Frames     : ^FrameArray;
                  Priority   : ^Priorities;
                  NumFrames  : Word;
                  NumPages   : Word;
                  FName      : PathStr;
                  F          : File;
                  PageLength : Word;  {Number of ELEMENTS in a Page}
                  LowestPty  : Word;  {Index of Frame with Lowest Priority}
                  MaxPty     : Real;  {Current Greatest Priority}
                  CurFrame   : Word;  {Index of current in-use Frame}

                  Procedure Init (NumElements : LongInt; ElementSize : Word;
                                  MaxBuffSize : LongInt; FileName : PathStr);

                  Function Load (ElementSize : Word; MaxBuffSize : LongInt;
                                  FileName : String) : LongInt;

                  Procedure Destroy;  {Erases File}
                  Procedure Release;  {Does Not Erase File}

                  Procedure Retrieve (Index : LongInt; Size : Word; Var El);
                  Procedure Accept (Index : LongInt; Size : Word; Var El);

                  Procedure UpDate_Priorities;
                  Procedure Reset_LowestPty;

                  Function Loaded_At (PageNumber : Word) : Word; {Return Frame}
                  Function Get_Page (Index : LongInt) : Word;  {Return Page}
                  Function Get_Frame (Index : LongInt) : Word; {Return Frame}
                  Procedure Swap_In (PageNumber : Word); {CurFrame = Frame}

                End;

IMPLEMENTATION

Const
  Infinity : Real = 1000000.0; {1 Million should do....}

Type
  Space = Array[0..0] of Byte;

Procedure Error (Num : Byte; Message : String);
Begin
  WriteLn;
  Write ('Page Manager Error [',Num,'] : ');
  Case Num of
          0 : WriteLn (' Unable to Open File "',Message,'". ');
          1 : WriteLn (' Attempted to create file larger then 32 MegaBytes in size. ');
          2 : WriteLn (' Attempted to Initialize with 0 elements. ');
          3 : WriteLn (' Unable to LOAD File "',Message,'" ');
          4 : WriteLn (' Attempted to LOAD with wrong Data Size. ');
       End;
  WriteLn;
  WriteLn ('***** PROGRAM TERMINATED *****');
  HALT (100+Num)
End;


Procedure PageManager.Init (NumElements : LongInt; ElementSize : Word;
                            MaxBuffSize : LongInt; FileName : PathStr);
Var
  Temp,I : Word;
  FSize  : LongInt;
  Dummy  : Page;
  FakeEl : ^Space;
  LTemp  : LongInt;
  L      : LongInt;

Begin
  MaxPty := 0.0;
  FName := FileName;
  If NumElements < 1 Then Error (2,'');
  FSize := NumElements * LongInt (ElementSize);
  NumPages := FSize Div Dummy.PageSize(ElementSize);
  If FSize Mod Dummy.PageSize(ElementSize) > 0 Then NumPages := NumPages + 1;

  If MaxBuffSize = 0 Then MaxBuffSize := MemAvail;
  If MaxBuffSize > (NumElements * LongInt(ElementSize)) +
                    LongInt(MaxFrame*SizeOf(Frame)+SizeOf(Real)*NumPages)
    Then
      MaxBuffSize := (NumElements * LongInt(ElementSize)) +
                      LongInt(MaxFrame*SizeOf(Frame)+SizeOf(Real)*NumPages);

  GetMem (Priority,NumPages*SizeOf(Real));
  For I := 0 to NumPages-1 do Priority^[I] := 0.0;

  If (NumPages * LongInt(Dummy.PageSize(ElementSize))) > MaximumSize
    Then
      Error (1,'');

  Temp := System.FileMode;
  System.FileMode := 2;     {Force Open to be for Read/Write}
  {$I-}
  Assign (F,FileName);
  Rewrite (F,1);
  {$I+}
  If IOResult <> 0 Then Error (0,FileName);
  System.FileMode := Temp;  {Put it back the way it was}

  Temp := Dummy.PageSize(ElementSize) + SizeOf(Frame);
  NumFrames := Word ((MaxBuffSize - NumPages*SizeOf(Real)) Div LongInt(Temp));
  If NumFrames = 0 Then NumFrames := 1;
  If NumFrames > NumPages Then NumFrames := NumPages;
  CurFrame := NumFrames+1;
  PageLength := Dummy.PageSize(ElementSize) Div ElementSize;

  GetMem (Frames,NumFrames*SizeOf(Frame));

  Temp := MaxPageSize Div ElementSize;
  If Temp > NumElements Then
    Begin
      Temp := NumElements;
      PageLength := Temp
    End;
  For I := 0 to NumFrames-1 do  {Initialize all frames}
    With Frames^[I] do
      Begin
        Priority^[I] := 0.0;
        PageNum := -1;
        Current.Init (Temp,ElementSize)
      End;

  LowestPty := 0;

  {The ONLY justification for the following block of code is}
  {the expediency of this method for initialization of the  }
  {the disk file.  I apologize to all the purists out there!}
  GetMem (FakeEl,ElementSize);
  For I := 0 to ElementSize-1 do FakeEl^[I] := 0;
  With Frames^[0] do
    Begin
      For L := 0 to Current.HighestEl-1 do
        Current.Accept (L,ElementSize,FakeEl^);
      For L := 0 to NumPages-1 do
        Begin                        {Fill File with Zeroes}
          LTemp := L*(Current.PageSize(ElementSize) Div ElementSize);
          Current.VirtAddr := LTemp;  {VERY,VERY *BAD* ! Don't do this!!!!!}
          While Current.HighestEl >= NumElements do Dec (Current.VirtAddr);
                                     {and ESPECIALLY NEVER do this!!!!!!!!}
          If LTemp < 0 Then {even worse....}
            Begin
              LTemp := 0;
              Current.VirtAddr := 0
            End;
          Current.Accept (LTemp,ElementSize,FakeEl^);
          Current.Flush (F)
        End;
      Current.VirtAddr := -1      {Undo the damage!}
    End;
  FreeMem (FakeEl,ElementSize);
  {End of *BAD* stuff!}

  Temp := 0;
  For I := 0 to NumFrames-1 do {Load the first NUMFRAMES Pages}
  With Frames^[I] do
    Begin
      PageNum := I;
      Current.Load (F,Temp);
      Temp := Temp+PageLength
    End;

End;

Function PageManager.Load;
Var
  Temp,I      : Word;
  FSize       : LongInt;
  Dummy       : Page;
  NumElements : LongInt;

Begin
  MaxPty := 0.0;
  FName := FileName;
  Temp := System.FileMode;
  System.FileMode := 2;     {Force Open to be for Read/Write}
  {$I-}
  Assign (F,FileName);
  Reset (F,1);
  {$I+}
  If IOResult <> 0 Then Error (3,FileName);
  System.FileMode := Temp;  {Put it back the way it was}
  FSize := FileSize (F);

  NumElements := FSize Div ElementSize;
  If FSize Mod NumElements <> 0 Then Error (4,'');

  NumPages := FSize Div Dummy.PageSize(ElementSize);
  If FSize Mod Dummy.PageSize(ElementSize) > 0 Then NumPages := NumPages + 1;

  If MaxBuffSize = 0 Then MaxBuffSize := MemAvail;
  If MaxBuffSize > (NumElements * LongInt(ElementSize)) +
                    LongInt(MaxFrame*SizeOf(Frame)+SizeOf(Real)*NumPages)
    Then
      MaxBuffSize := (NumElements * LongInt(ElementSize)) +
                      LongInt(MaxFrame*SizeOf(Frame)+SizeOf(Real)*NumPages);

  GetMem (Priority,NumPages*SizeOf(Real));
  For I := 0 to NumPages-1 do Priority^[I] := 0.0;

  Temp := Dummy.PageSize(ElementSize) + SizeOf(Frame);
  NumFrames := Word ((MaxBuffSize - NumPages*SizeOf(Real)) Div LongInt(Temp));
  If NumFrames = 0 Then NumFrames := 1;
  If NumFrames > NumPages Then NumFrames := NumPages;
  CurFrame := NumFrames+1;
  PageLength := Dummy.PageSize(ElementSize) Div ElementSize;

  GetMem (Frames,NumFrames*SizeOf(Frame));

  Temp := MaxPageSize Div ElementSize;
  If Temp > NumElements Then
    Begin
      Temp := NumElements;
      PageLength := Temp
    End;
  For I := 0 to NumFrames-1 do  {Initialize all frames}
    With Frames^[I] do
      Begin
        PageNum := -1;
        Current.Init (Temp,ElementSize)
      End;

  LowestPty := 0;

  Temp := 0;
  For I := 0 to NumFrames-1 do {Load the first NUMFRAMES Pages}
  With Frames^[I] do
    Begin
      PageNum := I;
      Current.Load (F,Temp);
      Temp := Temp+PageLength
    End;
  Load := NumElements
End;

Procedure PageManager.Destroy;
Var
  I : Word;
Begin
  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      Begin
        Current.Destroy;
        Priority^[I] := 0.0
      End;
  MaxPty := 0.0;
  FreeMem (Priority,NumPages*SizeOf(Real));
  FreeMem (Frames,NumFrames*SizeOf(Frame));
  Close (F);
  Erase (F)
End;

Procedure PageManager.Release;
Var
  I : Word;
Begin
  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      Begin
        Current.Flush (F);
        Current.Destroy;
        Priority^[I] := 0.0
      End;
  MaxPty := 0.0;
  FreeMem (Priority,NumPages*SizeOf(Real));
  FreeMem (Frames,NumFrames*SizeOf(Frame));
  Close (F);
End;

Procedure PageManager.UpDate_Priorities;
Var
  I,J  : Word;
  High : Real;
  Low  : Real;
Begin
  High := MaxPty;
  Low  := Infinity;

  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      Begin
        If Priority^[I] = Infinity Then Priority^[I] := 0.0; {give UP!}
        If Priority^[I] < Low Then
          Begin
            Low := Priority^[I];
            J := I
          End;
        If Priority^[I] > High Then High := Priority^[I]
      End;

  MaxPty := High
End;

Procedure PageManager.Reset_LowestPty;
Var
  I,J : Word;
  Low : Real;
Begin
  Low := Infinity;
  For I := 0 to NumFrames-1 do
    If Priority^[I] < Low Then
      Begin
        Low := Priority^[I];
        J := I
      End;
  LowestPty := J
End;

Function PageManager.Get_Page (Index : LongInt) : Word;
Begin
  Get_Page := Index Div PageLength
End;

Function PageManager.Loaded_At (PageNumber : Word) : Word;
Var
  I,J  : Word;
  Done : Boolean;
Begin
  J := NumFrames+1;         {Flags NOT Loaded}
  Done := False;
  I := 0;
  While (I <= NumFrames-1) and (Not Done) do
    Begin
      With Frames^[I] do
        If PageNum = PageNumber Then
          Begin
            Done := True;
            J := I
          End;
      Inc (I)
    End;
  Loaded_At := J
End;

Function PageManager.Get_Frame (Index : LongInt) : Word;
Begin
  Get_Frame := Loaded_At (Get_Page (Index))
End;

Procedure PageManager.Swap_In (PageNumber : Word);
Begin
  UpDate_Priorities;
  Reset_LowestPty;
  With Frames^[LowestPty] do
    Begin
      Current.Flush (F);
      Current.Load (F,(PageNumber)*PageLength);
      PageNum := PageNumber
    End;
  CurFrame := LowestPty
End;

Procedure PageManager.Accept (Index : LongInt; Size : Word; Var El);
Begin
  CurFrame := Get_Frame (Index);
  If CurFrame = NumFrames+1 Then Swap_In (Get_Page (Index));

  With Frames^[CurFrame] do
    Begin
      Current.Accept (Index,Size,El);
      Priority^[PageNum] := Priority^[PageNum] + Delta
    End
End;

Procedure PageManager.Retrieve (Index : LongInt; Size : Word; Var El);
Begin
  CurFrame := Get_Frame (Index);
  If CurFrame = NumFrames+1 Then Swap_In (Get_Page (Index));

  With Frames^[CurFrame] do
    Begin
      Current.Retrieve (Index,Size,El);
      Priority^[PageNum] := Priority^[PageNum] + Delta
    End
End;

{$F+}
Function HeapErrorTrap (Size : Word) : Integer;
Begin
  HeapErrorTrap := 1  { New and GetMem return Nil if out_of_memory }
End;
{$F-}

BEGIN
  HeapError := @HeapErrorTrap
END.
